import argparse
import sys

from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, Optional

from pandiag import dot, drawio, mermaid
from pandiag.model.graph import Graph
from pandiag.utils import help_escape

PLACEHOLDER = '%'

@dataclass
class Format:
    extensions: list[str] = field(default_factory=list)
    format: Optional[Callable[[Graph], str | bytes]] = None
    parse: Optional[Callable[[str], list[Graph]]] = None

FORMATS = {
    'dot': Format(extensions=['dot'], format=dot.format),
    'drawio': Format(extensions=['drawio'], parse=drawio.parse),
    'markdown': Format(extensions=['md', 'markdown'], format=mermaid.format_markdown),
    'mermaid': Format(extensions=[], format=mermaid.format),
    'pdf': Format(extensions=['pdf'], format=dot.format_pdf),
    'png': Format(extensions=['png'], format=dot.format_png),
}

def guess_format(path: Path) -> Optional[Format]:
    extension = path.name.rsplit('.', maxsplit=1)[-1]
    return next((f for f in FORMATS.values() if extension in f.extensions), None)

def main():
    parser = argparse.ArgumentParser(description='Utility for converting between diagram formats')
    parser.add_argument('-o', '--output', type=Path, required=True, metavar='PATH', help=f'The path to the output document. The file name must contain a {help_escape(PLACEHOLDER)} placeholder if the input document contains multiple graphs.')
    parser.add_argument('input', type=Path, help='The path to the input document.')

    args = parser.parse_args()
    input_path: Path = args.input
    output_template: Path = args.output
    
    if not (input_format := guess_format(input_path)):
        print(f'Unrecognized input format: {input_path.name}')
        sys.exit(1)
    
    if not (output_format := guess_format(output_template)):
        print(f'Unrecognized output format: {output_template.name}')
        sys.exit(1)

    if not input_format.parse:
        print('Input format does not support parsing')
        sys.exit(1)

    if not output_format.format:
        print('Output format does not support formatting')
        sys.exit(1)

    with open(input_path, 'r') as f:
        raw_input = f.read()

    graphs = input_format.parse(raw_input)

    if len(graphs) == 0:
        print('Input contains no graphs')
        sys.exit(1)

    if len(graphs) > 1 and (PLACEHOLDER not in output_template.stem):
        print(f'Output file name must contain a placeholder {PLACEHOLDER} since the input contains multiple ({len(graphs)}) graphs')
        sys.exit(1)

    for i, graph in enumerate(graphs):
        raw_output = output_format.format(graph)
        is_binary = isinstance(raw_output, bytes)

        if len(graphs) > 1:
            output_path = output_template.with_stem(output_template.stem.replace(PLACEHOLDER, f'{i:02d}'))
        else:
            output_path = output_template

        with open(output_path, 'wb' if is_binary else 'w') as f:
            f.write(raw_output)



from typing import Optional
from pandiag.model.graph import Edge, EdgeStyle, Graph, Node, Subgraph
from pandiag.utils import indent

import hashlib
import re

def _sanitize_label(label: str) -> str:
    # TODO: Find a better way to sanitize the label
    return re.sub(r'[()]', '', label.replace('\n', '<br/>')).strip() if label else None

def _format_label(label: Optional[str]) -> str:
    # TODO: Generate more readable node names than SHA1 hashes
    label = _sanitize_label(label)
    return f"{hashlib.sha1(label.encode()).hexdigest()}[{label}]" if label else 'None'

def _format_node(node: Node) -> str:
    return _format_label(node.label)

def _format_arrow(edge: Edge, graph: Graph) -> str:
    if graph.directed:
        match edge.style:
            case EdgeStyle.DASHED:
                return f"{'<-' if edge.source_arrow else '-'}.{'->' if edge.dest_arrow else '-'}"
            case EdgeStyle.SOLID:
                return f"{'<' if edge.source_arrow else '-'}-{'>' if edge.dest_arrow else '-'}"
    else:
        return '---'

def _format_edge(edge: Edge, graph: Graph) -> str:
    return f"{_format_label(edge.source)} {_format_arrow(edge, graph=graph)}{f'|{_sanitize_label(edge.label)}|' if edge.label else ''} {_format_label(edge.dest)}"

def _format_subgraph(subgraph: Subgraph, graph: Graph) -> list[str]:
    return [
        *(_format_node(n) for n in subgraph.nodes),
        *(_format_edge(e, graph=graph) for e in subgraph.edges),
        *(l for g in subgraph.subgraphs for l in [
            f'subgraph {_format_label(g.name)}',
            *indent(_format_subgraph(g, graph)),
            'end',
        ]),
    ]

def _format_graph(graph: Graph) -> list[str]:
    return [
        'flowchart TD',
        *indent(_format_subgraph(graph.rootgraph, graph=graph)),
    ]

def format(graph: Graph) -> str:
    return '\n'.join(_format_graph(graph))

def format_markdown(graph: Graph) -> str:
    return '\n'.join([
        '```mermaid',
        format(graph),
        '```',
    ])



from typing import Optional
from pandiag.model.graph import Edge, EdgeStyle, Graph, Node, Subgraph
from pandiag.utils import indent

import subprocess

def _format_label(label: Optional[str]) -> str:
    return f'"{label.replace('\n', '\\n').strip()}"' if label else None

def _format_node(node: Node) -> str:
    return f'{_format_label(node.label)};'

def _format_props(edge: Edge) -> str:
    props = {
        **(
            {'dir': 'both'} if edge.source_arrow and edge.dest_arrow else
            {'dir': 'back'} if edge.source_arrow and not edge.dest_arrow else
            {'dir': 'none'} if not edge.source_arrow and not edge.dest_arrow else
            {} # forward is the default direction, so we omit it for brevity
        ),
        **({'style': 'dashed'} if edge.style == EdgeStyle.DASHED else {}),
        **({'label': edge.label} if edge.label else {}),
    }
    return f"[{', '.join(f'{k}={v}' for k, v in props.items())}]" if props else ''

def _format_arrow(edge: Edge, graph: Graph) -> str:
    return '->' if graph.directed else '--'

def _format_edge(edge: Edge, graph: Graph) -> str:
    return f"{_format_label(edge.source)} {_format_arrow(edge, graph=graph)} {_format_label(edge.dest)} {_format_props(edge)};"

def _format_subgraph(subgraph: Subgraph, graph: Graph) -> list[str]:
    return [
        *(_format_node(n) for n in subgraph.nodes),
        *(_format_edge(e, graph=graph) for e in subgraph.edges),
        *(l for g in subgraph.subgraphs for l in [
            f"subgraph {_format_label(f'cluster_{g.name}')} {{",
            *indent([
                *([f'label = {_format_label(g.name)};'] if g.name else ''),
                *_format_subgraph(g, graph),
            ]),
            '}',
        ]),
    ]

def _format_graph(graph: Graph) -> list[str]:
    return [
        f"{'digraph' if graph.directed else 'graph'} {_format_label(graph.rootgraph.name)} {{",
        *indent(_format_subgraph(graph.rootgraph, graph=graph)),
        '}',
    ]

def format(graph: Graph) -> str:
    return '\n'.join(_format_graph(graph))

def format_rendered(graph: Graph, render_format: str) -> bytes:
    result = subprocess.run(
        ['dot', '-T', render_format],
        input=format(graph).encode(),
        capture_output=True
    )
    return result.stdout

def format_png(graph: Graph) -> bytes:
    return format_rendered(graph, 'png')

def format_pdf(graph: Graph) -> bytes:
    return format_rendered(graph, 'pdf')



from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

@dataclass
class Node:
    label: str

class EdgeStyle(Enum):
    SOLID = 1
    DASHED = 2

@dataclass
class Edge:
    source: Optional[str] = None
    dest: Optional[str] = None
    source_arrow: bool = False
    dest_arrow: bool = True
    style: EdgeStyle = EdgeStyle.SOLID
    label: Optional[str] = None

@dataclass
class Subgraph:
    name: Optional[str] = None
    nodes: list[Node] = field(default_factory=list)
    edges: list[Edge] = field(default_factory=list)
    subgraphs: list[Subgraph] = field(default_factory=list)

    def merge(self, other: Subgraph):
        self.nodes += other.nodes
        self.edges += other.edges
        self.subgraphs += other.subgraphs

@dataclass
class Graph:
    directed: bool = False
    rootgraph: Subgraph = field(default_factory=Subgraph)
